"""General utils"""

import inspect
from functools import partial

from py2http.decorators import mk_flat


def get_class_that_defined_method(meth):
    """Get the class from a method function object"""
    if isinstance(meth, partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (
            inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__,
                                                                                                '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def flat_callable_for(func, func_name=None, cls=None):
    """
    Flatten a simple cls->instance->method call pipeline into one function,
    or just return the input func (possibly with the func_name changed), if the function is already flat.

    That is, a function mk_flat(cls, method) that returns a "flat function" such that
    ```
    cls(**init_kwargs).method(**method_kwargs) == flat_func(**init_kwargs, **method_kwargs)
    ```

    :param method: The method object to get a flat function from
    :param func_name: The name to give the function
    :param cls: 
    :return: A flat function
    
    
    """
    containing_cls = get_class_that_defined_method(func)
    if not containing_cls:  # then it's just a "normal" function, so return it:
        if func_name is not None:
            func.__name__ = func_name
        return func
    else:
        if cls is None:
            cls = containing_cls
        if func_name is None:
            func_name = getattr(func, '__name__', 'flat_func')

        func = mk_flat(cls=cls, method=func, func_name=func_name)
        func.__qualname__ = getattr(func, '__qualname__', None)
        return func



import http2py
from py2http.service import run_http_service
from py2http.decorators import mk_flat, handle_json_req

from qh.trans import (
    transform_mapping_vals_with_name_func_map,
    mk_json_handler_from_name_mapping
)
from qh.util import (
    flat_callable_for
)
from qh.main import (
    mk_http_service_app
)


from typing import Iterable, Callable

from py2http import mk_http_service

from qh.util import flat_callable_for


def _ascertain_flat_func_list(funcs):
    if isinstance(funcs, Callable):
        return _ascertain_flat_func_list([funcs])
    else:
        assert isinstance(funcs, Iterable) and all(map(callable, funcs)), \
            f"funcs is supposed to be a callable or an iterable of callables: {funcs}"
        return [flat_callable_for(func) for func in funcs]


def mk_http_service_app(funcs, input_trans=None):
    funcs = _ascertain_flat_func_list(funcs)
    if input_trans:
        for func in funcs:
            func.input_mapper = input_trans
    return mk_http_service(funcs)




from typing import Mapping, Optional, Callable, Iterable
from py2http.decorators import mk_flat, handle_json_req

def _name_func_relationships_to_name_func_pairs(name_func_relationships):
    for k, v in name_func_relationships.items():
        if isinstance(k, str):
            name = k
            if isinstance(v, Callable):
                yield name, v
            elif isinstance(v, Iterable):
                for func in v:
                    assert isinstance(func, Callable), f"Should have been a callable: {func}"
                    yield name, func
            else:
                raise TypeError(f"value should have been a callable or an iterable of callables: {v}")
        elif isinstance(k, Callable):
            func = k
            if isinstance(v, str):
                yield v, func
            elif isinstance(v, Iterable):
                for name in v:
                    assert isinstance(name, str), f"Should have been a string: {name}"
                    yield name, func
            else:
                raise TypeError(f"value should have been a string or an iterable of strings: {v}")
        else:
            raise TypeError(f"key should have been a string or a callable: {k}")


def _name_func_relationships_to_name_func_map(name_func_relationships) -> dict:
    assert isinstance(name_func_relationships, Mapping), \
        f"name_func_relationships should be a Mapping of `name: func(s)` or `func: name(s)` pairs: " \
        f"{name_func_relationships}"
    name_func_items = list(_name_func_relationships_to_name_func_pairs(name_func_relationships))
    n = len(name_func_items)
    name_func_map = dict(name_func_items)
    assert len(name_func_map) == n, f"There were some duplicate names in name_func_relationships"
    return name_func_map


def transform_mapping_vals_with_name_func_map(mapping, name_func_map: Mapping[str, Callable]) -> dict:
    for name, val in mapping.items():
        if name in name_func_map:
            yield name, name_func_map[name](val)
        else:
            yield name, val


def mk_json_handler_from_name_mapping(name_func_relationships: Optional[Mapping] = None) -> dict:
    if name_func_relationships is None:
        @handle_json_req  # extracts the JSON body and passes it to the input mapper as a dict
        def input_trans(input_kwargs):
            return dict(input_kwargs)
    else:
        name_func_map = _name_func_relationships_to_name_func_map(name_func_relationships)

        @handle_json_req  # extracts the JSON body and passes it to the input mapper as a dict
        def input_trans(input_kwargs):
            return dict(transform_mapping_vals_with_name_func_map(input_kwargs, name_func_map))

        return input_trans
